"""Time series transformation and classification module for Python.

pyts is a Python package for time series classification.
It aims to make time series classification easily accessible by providing
preprocessing and utility tools, and implementations of state-of-the-art
algorithms. Most of these algorithms transform time series, thus pyts provides
several tools to perform these transformations.
"""

__version__ = '0.12.0'

__all__ = ['approximation', 'bag_of_words', 'classification', 'datasets',
           'decomposition', 'image', 'metrics', 'multivariate',
           'preprocessing', 'transformation', 'utils']
